# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
import matplotlib.pyplot as plot
import numpy as np

import os


def create_dir(dir_name: str):
    if dir_name is None or dir_name == "":
        return

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)


class GraphView:
    def __init__(self):
        self.sub_graphs = []

        self.graph_to_draw = nx.MultiDiGraph()

    def draw(self, title: (str, str) = None, show=True, output_dir: str = None, output_file: str = None,
             to_print=False):
        plot.close()

        if to_print:
            plot.figure(figsize=(10, 10), dpi=300)
        else:
            plot.figure(figsize=(5, 5), dpi=100)

        graph = self.graph_to_draw.copy()

        for (subgraph, color) in self.sub_graphs:
            graph = nx.compose(graph, subgraph)

        positions = nx.circular_layout(graph)

        for (subgraph, color) in self.sub_graphs:
            sg = graph.subgraph(subgraph.nodes())

            if to_print:
                nx.draw_networkx(sg, node_size=1000, font_size=20, with_labels=True,
                                 pos=positions, node_color=color, width=5.5, connectionstyle="arc3, rad = 0.1")
            else:
                nx.draw_networkx(sg, with_labels=True,
                                 pos=positions, node_color=color, connectionstyle="arc3, rad = 0.1")

            if title is not None:
                if title[0] is not None:
                    plot.title(title[0], fontsize=20)
                if title[1] is not None:
                    plot.xlabel(title[1], fontsize=20)

        if show:
            plot.show()

        if output_file is not None and output_dir is not None:
            create_dir(output_dir)
            plot.box(False)
            plot.savefig(os.path.join(output_dir, output_file))

    def graph_from_matrix(self, matrix: np.array, add_empty_nodes=True, color: str = "yellow", to_print=True):
        graph = self.graph_to_draw.copy()
        it = np.nditer(matrix, flags=["multi_index"])
        if to_print:
            renumber_nodes = 1
        else:
            renumber_nodes = 0

        for edges in it:
            if edges > 0:
                for i in range(edges):
                    graph.add_edge(it.multi_index[0] + renumber_nodes, it.multi_index[1] + renumber_nodes)
            else:
                if add_empty_nodes:
                    graph.add_node(it.multi_index[0] + renumber_nodes)

        self.sub_graphs.append((graph, color))

    def graph_from_edges(self, edges: list[list[int]], color: str = "yellow"):
        graph = self.graph_to_draw.copy()

        for edge in edges:
            graph.add_edge(edge[0], edge[1])

        self.sub_graphs.append((graph, color))

    def clear(self):
        self.sub_graphs = []



import numpy as np


class SequenceGenerator:
    def __init__(self, sequence_length, sequence_min: int, sequence_max: int, dtype=np.uint16, seed=None):
        self.dtype = dtype

        self.sequence_length = sequence_length
        self.sequence_min = sequence_min
        self.sequence_max = sequence_max

        self.unique_sequences = np.empty((0, self.sequence_length), dtype=self.dtype)
        if seed is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = np.random.default_rng(seed=seed)

    def __generate_new_sequences(self, number_sequences: int = 1, unique_elements=False) -> np.ndarray:
        rng = self.rng

        if unique_elements:
            sequences = np.zeros((number_sequences, self.sequence_length), dtype=self.dtype)
            for i in range(number_sequences):
                sequences[i] = rng.choice(self.sequence_max - self.sequence_min,
                                          size=self.sequence_length, replace=False) + self.sequence_min
        else:
            sequences = rng.integers(self.sequence_min, self.sequence_max,
                                     size=(number_sequences, self.sequence_length), dtype=self.dtype)
        return sequences

    def generate_unique_sequences(self, number_sequences=1, unique_elements=False) -> np.ndarray:
        number_to_add = number_sequences
        number_added = 0
        added_sequences = np.empty((0, self.sequence_length), dtype=self.dtype)

        while number_added < number_sequences:
            sequences = self.__generate_new_sequences(number_sequences=number_to_add, unique_elements=unique_elements)
            unique_sequences = np.unique(sequences, axis=0)
            added_sequences = np.unique(np.append(added_sequences, unique_sequences, axis=0), axis=0)
            number_added = len(added_sequences)
            number_to_add = number_sequences - number_added

        if unique_elements:
            self.unique_sequences = np.append(self.unique_sequences, added_sequences, axis=0)

        return added_sequences

    @staticmethod
    def intersect_2d(array_1: np.ndarray, array_2: np.ndarray) -> np.ndarray:
        n_cols = array_1.shape[1]
        dtype = {'names': ['f{}'.format(i) for i in range(n_cols)],
                 'formats': n_cols * [array_1.dtype]}

        intersection = np.intersect1d(array_1.view(dtype), array_2.view(dtype))
        intersection = intersection.view(array_1.dtype).reshape(-1, n_cols)

        return intersection

    @staticmethod
    def setdiff_2d(array_1: np.ndarray, array_2: np.ndarray) -> np.ndarray:
        n_cols = array_1.shape[1]
        dtype = {'names': ['f{}'.format(i) for i in range(n_cols)],
                 'formats': n_cols * [array_1.dtype]}

        difference = np.setdiff1d(array_1.view(dtype), array_2.view(dtype))
        difference = difference.view(array_1.dtype).reshape(-1, n_cols)

        return difference

    def add_unique_sequences(self, number_sequences=1) -> (np.ndarray, np.ndarray):
        number_to_add = number_sequences
        number_added = 0

        init_unique_sequences = len(self.unique_sequences)
        all_sequences = np.copy(self.unique_sequences)

        unique_added = np.empty((0, self.sequence_length), dtype=self.dtype)

        while number_added < number_sequences:
            sequences = self.generate_unique_sequences(number_sequences=number_to_add, unique_elements=True)

            intersection = self.intersect_2d(all_sequences, sequences)

            unique_to_add = self.setdiff_2d(sequences, intersection)
            unique_added = np.append(unique_added, unique_to_add, axis=0)

            unique_sequences, indexes = np.unique(np.append(all_sequences, sequences, axis=0), axis=0,
                                                  return_index=True)

            all_sequences = unique_sequences.copy()

            number_added = len(all_sequences) - init_unique_sequences
            number_to_add = number_sequences - number_added

        self.unique_sequences = all_sequences

        return unique_added, self.unique_sequences



# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plot
import numpy as np
import pandas as pd
import seaborn as sns
import warnings
import sys

from pyprind import ProgBar

from ssakg.ordering_algorithms import SimpleSort, NodeOrderingAlgorithm, EnhancedNodeOrderingAlgorithm, \
    WeightedEdgesNodeOrderingAlgorithm, OrderingAlgorithm
from ssakg.ssakg import SSAKG


class SSAKG_Tester:
    def __init__(self, ssakg: SSAKG, sequences: np.ndarray, algorithms_list: list[OrderingAlgorithm] = None):
        self.ssakg = ssakg
        self.sequences = sequences
        self.sequence_length = len(sequences[0])
        self.algorithms_test = {}
        self.context_length = None
        self.title = ""
        self.x_label = "Number of correct elements in sequence"
        self.y_label = "Number of sequences"
        self.algorithms_list = algorithms_list
        self.unsorted_elements_test = [0, 0]
        if algorithms_list is None:
            self.algorithms_list = [SimpleSort(), NodeOrderingAlgorithm(),
                                    EnhancedNodeOrderingAlgorithm(),
                                    WeightedEdgesNodeOrderingAlgorithm()]

    def add_algorithm(self, name: str):
        histogram = np.zeros(self.sequence_length + 1, dtype=np.int16)
        corrections = 0
        all_values = 0
        self.algorithms_test[name] = [histogram, corrections, all_values]

    def ordering_test(self, translated_sequence, context_sequence):
        use_only_first_path = True
        for algorithm in self.algorithms_list:
            sorted_sequence, _ = self.ssakg.order_sequence(context_sequence, algorithm,
                                                           use_only_first_path)
            result, agreement = SSAKG.compare_sorted_sequences(translated_sequence, sorted_sequence)
            self.add_values(agreement, str(algorithm))

    def add_value(self, algorithm_name: str, agreement: np.ndarray):
        no_result = np.sum(agreement)
        histogram = self.algorithms_test[algorithm_name][0]

        histogram[no_result] += 1

        if no_result >= self.sequence_length:
            self.algorithms_test[algorithm_name][1] += 1

        self.algorithms_test[algorithm_name][2] += 1

    def add_values(self, agreement: np.ndarray, name: str):
        if name in self.algorithms_test:
            self.add_value(name, agreement)
        else:
            self.add_algorithm(name)
            self.add_value(name, agreement)

    def clear(self):
        self.algorithms_test = {}
        self.unsorted_elements_test = [0, 0]

    def _translate_sequence(self, sequence: np.ndarray) -> np.ndarray:
        return self.ssakg.translate_sequence(sequence)

    def _create_context(self, sequence: np.ndarray, context_length: int) -> np.ndarray:
        translated_sequence = self._translate_sequence(sequence)
        context = SSAKG.context_from_sequence(context_length, translated_sequence)

        return context

    def _sequence_to_compare_elements(self, sequence: np.ndarray) -> np.ndarray:
        return self._translate_sequence(sequence)

    def _sequence_to_ordering_test(self, sequence: np.ndarray) -> np.ndarray:
        return self._translate_sequence(sequence)

    def _sequence_with_unsorted_elements(self, sequence: np.ndarray) -> np.ndarray:
        return sequence

    def make_test(self, context_length, show_progress=False):
        self.clear()
        self.context_length = context_length

        if context_length > self.sequence_length:
            warnings.warn("Context is longer than sequence length, unable to make test")

        if show_progress:
            bar = ProgBar(len(self.sequences), stream=sys.stdout, title="ssakg test progress")
        else:
            bar = None

        # The context may contain recurring elements. Symbols assigned to such elements are assigned dynamically.
        # To obtain a context with recurring elements,
        # the simplest way is to translate the entire sequence and choose the desired context.

        for i in range(len(self.sequences)):
            context = self._create_context(self.sequences[i], context_length)
            # Now we have correct context in a direct way. We can read unsorted elements from ssakg.
            # Context is currently translated
            unsorted_sequence, _ = self.ssakg.get_unsorted_elements(context, context_is_translated=True)

            if SSAKG.compare_sets_of_elements(self._sequence_to_compare_elements(self.sequences[i]),
                                              self._sequence_with_unsorted_elements(unsorted_sequence)):
                self.unsorted_elements_test[0] += 1
                self.ordering_test(self._sequence_to_ordering_test(self.sequences[i]), unsorted_sequence)
            else:
                self.unsorted_elements_test[1] += 1

            if bar is not None:
                bar.update()

    def plot_agreement_histogram(self, draw_text: bool = True):
        self.title = f"\n Comparison of algorithms \n Sequence length {self.sequence_length}, context length {self.context_length}"

        word_number = []
        correct_no = []
        algorithms_name = []

        for algorithm_name, values_set in self.algorithms_test.items():
            for i, value in enumerate(values_set[0]):
                word_number.append(i)
                correct_no.append(value)
                algorithms_name.append(algorithm_name)

        dataframe = pd.DataFrame({"word_number": word_number, "correct_no": correct_no, "algorithm": algorithms_name})

        fig, ax = plot.subplots()

        bar_plot = sns.barplot(dataframe, x="word_number", y="correct_no", ax=ax, hue="algorithm", legend=True,
                               palette="bright", width=1)
        bar_plot.legend().set_title("")

        plot.grid()

        if draw_text:
            plot.title(
                self.title)
        plot.xlabel(self.x_label)
        plot.ylabel(self.y_label)

        plot.show()

    def create_dataframe(self):
        data = {}
        headers = []

        data["no sequences"] = []
        data["correct"] = []
        data["incorrect"] = []
        data["correct sort percentage"] = []

        for algorithm_name, values_set in self.algorithms_test.items():

            headers.append(algorithm_name)
            correct_sort = values_set[1]
            all_values = values_set[2]

            correction_sort_percentage = 0
            incorrect = 0

            if correct_sort > 0:
                correction_sort_percentage = correct_sort / all_values * 100
                incorrect = all_values - correct_sort

            data["no sequences"].append(all_values)
            data["correct"].append(correct_sort)
            data["incorrect"].append(incorrect)
            data["correct sort percentage"].append(correction_sort_percentage)

        df = pd.DataFrame(data,
                          index=headers)

        return df

    def print_dataframe(self):
        df = self.create_dataframe()
        print(df.to_string(formatters={"correct sort percentage": "{:2,.2f}%".format}))

    def __str__(self):
        algorithm_info = ""
        graph_matrix = self.ssakg.get_graph()
        algorithm_info += f"ssakg dimension: {len(graph_matrix)}\n"
        algorithm_info += f"sequence length: {self.sequence_length}\n"
        algorithm_info += f"context length: {self.context_length}\n"
        unsorted_percentage = (self.unsorted_elements_test[0] /
                               (self.unsorted_elements_test[0] + self.unsorted_elements_test[1]) * 100)
        algorithm_info += f"unordered sequences restored: {unsorted_percentage:.2f}%\n"
        df = self.create_dataframe()
        algorithm_info += df.to_string(formatters={"correct sort percentage": "{:2,.2f}%".format})

        return algorithm_info



# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np

import warnings

import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plot

from ssakg.anakg_view import GraphView, create_dir
from ssakg.sequence_translator import SequenceTranslator

from tabulate import tabulate


class ANAKG:
    def __init__(self, graph_dim: int = 10, subgraph_dim: int = 5, dtype=np.int16, graphs_to_drawing=False,
                 remove_diagonals=True, weighted_edges=True):

        # The parameter graphs_to_drawing is only for draw colorfully graph to examples.
        # Do not use it for other purposes.

        self.graph_dim = graph_dim
        self.subgraph_dim = subgraph_dim
        self.dtype = dtype

        self.sequence_translator = SequenceTranslator(sequence_matrix_dim=graph_dim, dtype=self.dtype)

        self.subgraph_pattern = self.create_upper_triangular(self.graph_dim, remove_diagonals=remove_diagonals,
                                                                     weighted_edges=weighted_edges, dtype=self.dtype)

        self.graph = self.crate_graph()
        self.sequences = []

        self.graphs_to_drawing = graphs_to_drawing
        self.graph_matrices = []

    @staticmethod
    def create_upper_triangular(dim: int, remove_diagonals=False, weighted_edges=True, dtype=np.int16) -> np.array:

        array = np.ones([dim, dim], dtype=dtype)

        if weighted_edges:
            mul = np.arange(1, dim + 1, dtype=dtype)
        else:
            mul = np.ones(dim, dtype=dtype)

        upper_triangular_matrix = (np.tril(array) * mul).transpose()

        if remove_diagonals:
            upper_triangular_matrix = upper_triangular_matrix - np.eye(dim, dtype=dtype) * mul

        return upper_triangular_matrix

    @staticmethod
    def create_square(dim: int, remove_diagonals=False, weighted_edges=False, dtype=np.int16) -> np.array:
        if weighted_edges:
            mul = np.arange(1, dim + 1, dtype=dtype)
        else:
            mul = np.ones(dim, dtype=dtype)

        square_matrix = (np.ones([dim, dim], dtype=dtype) * mul).transpose()

        if remove_diagonals:
            square_matrix = square_matrix - np.eye(dim, dtype=dtype) * mul

        return square_matrix

    def set_subgraph_pattern(self, subgraph_pattern: np.array):
        self.subgraph_pattern = subgraph_pattern

    def crate_graph(self):
        dim = self.graph_dim

        return np.zeros([dim, dim], dtype=self.dtype)

    def add_subgraph(self, i: int, j: int):
        subgraph = self.subgraph_pattern.copy()

        temp_graph = np.zeros(self.graph.shape, dtype=self.dtype)
        temp_graph[i:i + self.subgraph_dim, j:j + self.subgraph_dim] = subgraph
        if self.graphs_to_drawing:
            self.graph_matrices.append(temp_graph)

        self.graph = self.graph + temp_graph

    def add_random_subgraph(self):
        rand_min = 0
        rand_max = len(self.graph) - self.subgraph_dim + 1
        i = np.random.randint(rand_min, rand_max)
        j = np.random.randint(rand_min, rand_max)

        self.add_subgraph(i, j)

    def add_dim_to_graph(self, dim_to_add: int):
        self.graph_dim += dim_to_add
        self.graph = np.pad(self.graph, (0, dim_to_add), 'constant')

        for i in range(len(self.graph_matrices)):
            self.graph_matrices[i] = np.pad(self.graph_matrices[i], (0, dim_to_add), 'constant')

    def sequence_to_subgraph(self, sequence: np.ndarray) -> np.ndarray:
        temp_graph = np.zeros(self.graph.shape, dtype=self.dtype)

        for i in range(self.subgraph_dim):
            for j in range(self.subgraph_dim):
                temp_graph[sequence[i], sequence[j]] = self.subgraph_pattern[i, j]

        return temp_graph

    def insert_sequence(self, sequence: np.ndarray) -> np.ndarray | None:
        if len(sequence) != self.subgraph_dim:
            warnings.warn(f"Sequence has incorrect length {len(sequence)}. Subgraph not inserted.")
            return None

        new_sequence, dims_to_add = self.sequence_translator.translate_sequence(sequence)

        if dims_to_add > 0:
            self.add_dim_to_graph(dims_to_add)

        self.sequences.append(new_sequence)
        temp_graph = self.sequence_to_subgraph(new_sequence)

        if self.graphs_to_drawing:
            self.graph_matrices.append(temp_graph)

        self.graph = self.graph + temp_graph

        return new_sequence

    def insert(self, sequences: np.ndarray):
        if np.ndim(sequences) == 1:
            return self.insert_sequence(sequences)
        else:
            for sequence in sequences:
                self.insert_sequence(sequence)

    def get_subgraph_from_sequence(self, sequence: np.ndarray) -> np.ndarray | None:
        if len(sequence) > self.subgraph_dim:
            warn = f"Sequence has incorrect length {len(sequence)}. Subgraph not created."
            warnings.warn(warn)
            return None

        temp_graph = np.zeros([len(sequence), len(sequence)], dtype=self.dtype)

        for i in range(len(sequence)):
            for j in range(len(sequence)):
                temp_graph[i, j] = self.graph[sequence[i], sequence[j]]

        return temp_graph

    def __str__(self):
        graph_size = self.graph.nbytes
        sequences_no = len(self.graph_matrices)
        sequences_size = 0
        for sequence in self.graph_matrices:
            sequences_size += sequence.nbytes

        density = self.density()

        return (f"graph_size: {graph_size} B\ngraph density: {density:.2f}\nstored sequences: "
                f"{sequences_no}\nstored sequences size: {sequences_size}B\n")

    def get_subgraph(self):
        return self.subgraph_pattern

    def get_graph(self):
        return self.graph

    def create_graph_view(self, to_print=False):
        graph_view = GraphView()
        graph_view.graph_from_matrix(np.zeros(self.graph.shape), color="gray", to_print=to_print)

        last_matrix = np.zeros(self.graph.shape, dtype=self.dtype)

        for graph_matrix in self.graph_matrices:
            last_matrix += graph_matrix

        graph_view.graph_from_matrix(last_matrix, add_empty_nodes=False, color="orange", to_print=to_print)

        return graph_view

    def density(self) -> float:
        n = self.graph_dim
        non_zeros = np.nonzero(self.graph)
        m = len(non_zeros[0])

        return m / n / (n - 1)

    def show(self, to_print=False):
        if not self.graphs_to_drawing:
            warnings.warn("Sequences are not stored. Graph view is empty.")
            return

        graph_view = self.create_graph_view(to_print=to_print)
        graph_view.draw(title=None, to_print=to_print)

    def save_fig(self, title: (str, str), output_dir: str = None, output_file: str = None, to_print=True):
        if not self.graphs_to_drawing:
            warnings.warn("Sequences are not stored. Graph view is empty.")

        graph_view = self.create_graph_view(to_print=to_print)
        graph_view.draw(show=False, output_dir=output_dir, output_file=output_file, title=title, to_print=to_print)

    def save_matrix(self, output_dir: str = None, output_file: str = None):
        if not self.graphs_to_drawing:
            warnings.warn("Sequences are not stored. Matrix is empty.")

        if output_file is not None and output_dir is not None:
            create_dir(output_dir)
            file = open(os.path.join(output_dir, output_file), "w")
            file.write(tabulate(self.graph, tablefmt="latex", floatfmt=".0f"))
            file.close()

    def translate_sequence(self, sequence: np.ndarray):
        return self.sequence_translator.translate_sequence(sequence, add_to_dictionary=False)[0]

    def decode_sequence(self, sequence: np.ndarray):
        return self.sequence_translator.decode_sentence(sequence)

    def check_context_correctness(self, context: np.ndarray):
        pass

    def sequence_from_placeholders(self, sequence: list[str]) -> np.ndarray:
        return self.sequence_translator.sequence_from_placeholders(sequence)

    def translate_context(self, context: np.ndarray):
        return self.sequence_translator.translate_sequence(context, add_to_dictionary=False)[0]

    def clear(self):
        self.graph = self.crate_graph()
        self.sequences = []
        self.sequence_translator = SequenceTranslator(sequence_matrix_dim=self.graph_dim, dtype=self.dtype)
        self.graph_matrices = []

    def draw_density(self, title=None, x_label=None, y_label=None, circle_size: int = 2):
        dim = self.graph_dim

        graph_1d = self.graph.reshape(dim * dim)

        non_zeros = np.nonzero(graph_1d)[0]
        x = np.mod(non_zeros, dim)
        y = np.floor_divide(non_zeros, dim)
        non_zeros_values = graph_1d[non_zeros]

        dataframe = pd.DataFrame({"x": x, "y": y, "z": non_zeros_values})

        palette = sns.color_palette("dark:#5A9_r", as_cmap=True)
        joint_grid = sns.JointGrid(data=dataframe, x="x", y="y", hue="z", palette=palette)
        joint_grid.plot_joint(sns.scatterplot, s=circle_size)
        joint_grid.plot_marginals(sns.histplot, kde=False, bins=100)
        joint_grid.ax_joint.legend().remove()

        if title is not None:
            joint_grid.fig.suptitle("ssakg - local density")

        if x_label is not None:
            joint_grid.ax_joint.set_xlabel(x_label)
        else:
            joint_grid.ax_joint.set_xlabel("")

        if y_label is not None:
            joint_grid.ax_joint.set_ylabel(y_label)
        else:
            joint_grid.ax_joint.set_ylabel("")

        joint_grid.fig.tight_layout()

        plot.show()



# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import warnings

from ssakg.anakg import ANAKG
from ssakg.ordering_algorithms import OrderingAlgorithm, WeightedEdgesNodeOrderingAlgorithm


class SSAKG(ANAKG):
    def __init__(self, number_of_symbols: int = 10, sequence_length: int = 5, dtype=np.int16, graphs_to_drawing=False,
                 remove_diagonals=True, weighted_edges=True):
        super().__init__(number_of_symbols, sequence_length, dtype, graphs_to_drawing, remove_diagonals,
                         weighted_edges)
        self.new_sequences_added = False
        self.symmetric_array_mask = self.eval_symmetric_array()

    def set_array(self, graph_array: np.ndarray):
        self.symmetric_array_mask = graph_array + np.transpose(graph_array)

    def eval_symmetric_array(self):
        self.symmetric_array_mask = self.get_graph() + np.transpose(self.get_graph())
        self.symmetric_array_mask = self.symmetric_array_mask + np.eye(self.symmetric_array_mask.shape[0])
        self.symmetric_array_mask = np.where(self.symmetric_array_mask == 0, 0, 1)
        self.new_sequences_added = False
        return self.symmetric_array_mask

    def get_unsorted_elements(self, context: np.ndarray, context_is_translated=False) -> (
            np.ndarray, np.ndarray):
        if self.new_sequences_added:
            self.eval_symmetric_array()

        if context_is_translated:
            translated_context = context
            unique_context = np.unique(context)
            if len(unique_context) != len(context):
                warnings.warn("List of symbols contains duplicate elements. Please use underscored symbols instead e.g. \"5_2\".")
                return None, None
        else:
            translated_context = self.translate_context(context)

        if translated_context is None:
            return None, None

        context_array = self.symmetric_array_mask[translated_context]

        if len(context_array) == 0:
            return None

        row_prod = np.prod(context_array, axis=0)
        non_zeros = np.where(row_prod != 0)[0]

        return non_zeros, row_prod

    def __get_sequence(self, context: np.ndarray, decode_sequence=True, context_is_translated=False,
                       ordering_alg=WeightedEdgesNodeOrderingAlgorithm()) -> np.ndarray | list:

        unsorted_elements, _ = self.get_unsorted_elements(context, context_is_translated)
        if unsorted_elements is None:
            return None
        sorted_elements, _ = self.order_sequence(unsorted_elements, ordering_alg=ordering_alg,
                                                 use_only_first_path=True)

        if decode_sequence:
            return self.decode_sequence(sorted_elements)

        return sorted_elements

    def get_sequence(self, context):
        context = self.sequence_from_placeholders(context)
        return self.__get_sequence(context, decode_sequence=True, context_is_translated=True)

    def insert_sequence(self, sequence: np.ndarray) -> np.ndarray | None:
        self.new_sequences_added = True
        return super().insert_sequence(sequence)

    def insert(self, sequences: np.ndarray):
        self.new_sequences_added = True
        return super().insert(sequences)

    @staticmethod
    def context_from_sequence(context_length: int, sequence: np.ndarray) -> np.ndarray:
        indexes = np.random.permutation(len(sequence))
        indexes = indexes[:context_length]

        context_array = sequence[indexes]
        return context_array

    @staticmethod
    def compare_sorted_sequences(sequence_1: np.ndarray, sequence_2: np.ndarray) -> (bool, np.ndarray):
        agreement = np.zeros(len(sequence_1), dtype=np.int8)

        if sequence_1 is None or sequence_2 is None:
            return False, agreement

        if len(sequence_1) != len(sequence_2):
            return False, agreement

        for i in range(len(sequence_1)):
            if sequence_1[i] == sequence_2[i]:
                agreement[i] = 1

        return np.array_equal(sequence_1, sequence_2), agreement

    @staticmethod
    def compare_sets_of_elements(sequence_1: np.ndarray, sequence_2: np.ndarray) -> bool:
        if len(sequence_1) != len(sequence_2):
            return False
        else:
            sequence_1 = np.sort(sequence_1)
            sequence_2 = np.sort(sequence_2)

            if not np.array_equal(sequence_1, sequence_2):
                print("ssakg: different sequences")
                return False

            return True

    def order_sequence(self, sequence: np.ndarray, ordering_alg: OrderingAlgorithm | None = None,
                       use_only_first_path=False) -> (
            np.ndarray | None, int):
        sequence_matrix = self.get_subgraph_from_sequence(sequence)
        if sequence_matrix is None:
            return None, 0

        orderings = ordering_alg(sequence_array=sequence_matrix, use_only_first_path=use_only_first_path)

        new_ordering = orderings[0]
        sorted_sequence = sequence[new_ordering]

        return sorted_sequence, len(orderings)



from setuptools import setup, find_packages

VERSION = '0.0.1'
DESCRIPTION = 'Sequential Structural Associative Knowledge Graph (ssakg)'
LONG_DESCRIPTION = open('README_PYPI.md').read()

setup(
    name='ssakg',
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['numpy~=1.26.2', "networkx~=3.2.1", "matplotlib~=3.8.2", "pandas~=2.2.0", "tabulate~=0.9.0",
                      "PyPrind~=2.11.3", "seaborn~=0.13.2"],
    url='https://github.com/PrzemyslawStok/ssakg',
    license='Apache 2.0',
    author='Przemysław Stokłosa',
    author_email='przemyslaw.stoklosa@gmail.com',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords=["semantic memory", "structural graphs", "graph density", "sequence storage", "sequence retrieval",
              "context based memory", "graph based memory"],
    python_requires='>=3.10',
)



# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import re

import warnings


class SequenceTranslator:
    def __init__(self, sequence_matrix_dim: int, dtype=np.int16):
        self.base_sequence_length = sequence_matrix_dim
        self.sequence_matrix_dim = sequence_matrix_dim
        self.dtype = dtype
        self.sequence_symbols_dictionary = np.zeros((self.sequence_matrix_dim, self.sequence_matrix_dim),
                                                    dtype=self.dtype)
        self.sequence_tuples_dictionary = np.zeros((0, 2), dtype=self.dtype)

        self.sequence_symbols_counter = np.zeros([sequence_matrix_dim], dtype=self.dtype)
        self.max_sequence = self.__max_sequence_number()
        self.multiple_symbol_pattern = re.compile(r"^\s*(\d+)_(\d+)\s*$")

    def __max_sequence_number(self):
        return np.max(self.sequence_symbols_dictionary)

    def __symbol_number_from_tuple(self, symbol_tuple):
        # This function returns encoded number of multiple symbol.
        if symbol_tuple[1] == 1:
            return symbol_tuple[0]
        else:
            address_value = self.sequence_symbols_dictionary.item(symbol_tuple)
            return address_value + self.base_sequence_length - 1

    def __translate_sequence_tuples(self, sequence_tuple: np.ndarray) -> np.ndarray:
        # This function translates sequences with repeated elements into sequences with unique elements.

        new_sequence = np.zeros(len(sequence_tuple), dtype=self.dtype)
        for index, enumerated_sequence_tuple in enumerate(sequence_tuple):
            address_value = self.sequence_symbols_dictionary.item(tuple(enumerated_sequence_tuple))
            if address_value == 0:
                new_sequence[index] = enumerated_sequence_tuple[0]
            else:
                new_sequence[index] = address_value + self.base_sequence_length - 1

        return new_sequence

    def __sequence_to_enumerated_tuples(self, sequence: np.ndarray) -> np.ndarray:
        current_sequence_symbols_counter = np.copy(self.sequence_symbols_counter)

        # The first item of tuple stores symbol,
        # the second item stores number of occurrences of the symbol in the sequence.
        sequence_tuples = np.zeros([len(sequence), 2], dtype=self.dtype)

        for i in range(len(sequence)):
            current_sequence_symbols_counter[sequence[i]] += 1
            sequence_tuples[i] = (sequence[i], current_sequence_symbols_counter[sequence[i]])

        return sequence_tuples

    def translate_sequence(self, sequence: np.ndarray, add_to_dictionary=True) -> (np.ndarray, int):
        rows_to_add = 0
        sequence_tuples = self.__sequence_to_enumerated_tuples(sequence)

        for enumerated_sequence_item in sequence_tuples:
            if enumerated_sequence_item[1] > 1:
                sequence_address = tuple(enumerated_sequence_item)
                if self.sequence_symbols_dictionary.item(sequence_address) == 0:
                    if add_to_dictionary:
                        # The sequence contains doubled elements, not exists in sequences dictionary.
                        # This section adds doubled elements to the dictionary.
                        self.max_sequence += 1
                        self.sequence_symbols_dictionary[sequence_address] =  self.max_sequence
                        self.sequence_tuples_dictionary = np.append(self.sequence_tuples_dictionary,
                                                                    np.array([sequence_address]), axis=0)
                        rows_to_add += 1
                    else:
                        # This sequence contains doubled elements with are not exist in sequences dictionary.
                        # This kind of sequence is not possible to translate. Returning None
                        return None, 0

        new_sequence_tuple = self.__translate_sequence_tuples(sequence_tuples)

        return new_sequence_tuple, rows_to_add

    def decode_sentence(self, sequence: np.ndarray) -> np.ndarray:
        current_enumerated_sequence = np.zeros(np.shape(sequence), dtype=self.dtype)

        for (i, symbol) in enumerate(sequence):
            if symbol < len(self.sequence_symbols_dictionary):
                current_enumerated_sequence[i] = symbol
            else:
                symbol_index = symbol - len(self.sequence_symbols_dictionary)
                if symbol_index >= len(self.sequence_symbols_dictionary):
                    current_enumerated_sequence[i] = None
                else:
                    current_enumerated_sequence[i] = self.sequence_tuples_dictionary[symbol_index][0]

        return current_enumerated_sequence

    def __check_symbol_value_in_range(self, symbol: int) -> bool:
        if symbol > self.sequence_matrix_dim or symbol < 0:
            return False
        else:
            return True

    def __symbol_to_tuple(self, symbol: str) -> tuple[int, int] | None:
        result = re.match(self.multiple_symbol_pattern, symbol)
        if result is None or len(result.groups()) != 2:
            return None
        return int(result.group(1)), int(result.group(2))

    def sequence_from_placeholders(self, sequence: list[str]) -> np.ndarray:
        decoded_sequence = []
        for symbol in sequence:
            if type(symbol) == str:
                tuple_symbol = self.__symbol_to_tuple(symbol)
                if self.__check_symbol_value_in_range(tuple_symbol[0]):
                    symbol_number = self.__symbol_number_from_tuple(tuple_symbol)
                    decoded_sequence.append(symbol_number)
                else:
                    warnings.warn(f"Symbol {symbol} not in range, will be omitted.")
            else:
                if self.__check_symbol_value_in_range(int(symbol)):
                    decoded_sequence.append(symbol)
                else:
                    warnings.warn(f"Symbol {symbol} not in range, will be omitted.")

        return np.array(decoded_sequence)

    def __str__(self):
        return str(self.sequence_symbols_dictionary)


if __name__ == "__main__":
    sequence_translator = SequenceTranslator(11)

    print(sequence_translator.translate_sequence(np.array([3, 4, 3, 10])))
    print(sequence_translator.translate_sequence(np.array([1, 4, 1, 4])))
    print(sequence_translator.translate_sequence(np.array([5, 5, 0, 5])))
    print(sequence_translator.translate_sequence(np.array([2, 4, 1, 4])))
    print(sequence_translator)

    unique_sequence, _ = sequence_translator.translate_sequence(np.array([5, 5, 5, 5]), add_to_dictionary=True)
    print(unique_sequence)

    enumerated_sequence = sequence_translator.decode_sentence(unique_sequence)
    print(enumerated_sequence)



# Copyright 2024 The SSAKG Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class OrderingAlgorithm:
    def __init__(self, dtype=np.int16):
        self.dtype = dtype

    def __call__(self, sequence_array: np.ndarray, use_only_first_path=False):
        all_paths = self.__ordering_algorithm(sequence_array, [], [], max_index=len(sequence_array) + 1,
                                              use_only_first_path=use_only_first_path)
        new_paths = []
        for path in all_paths:
            new_paths.append(OrderingAlgorithm.ordering_to_matrix_indices(path))

        return new_paths

    def __str__(self):
        pass

    def _ordering_function(self, array: np.ndarray, max_index: int) -> np.ndarray:
        pass

    @staticmethod
    def find_indexes_with_min_zeros_count(array: np.ndarray):
        number_of_zeros_in_row = np.sum(array == 0, axis=1)
        min_zeros = np.min(number_of_zeros_in_row)

        indices_with_min_zeros_values = np.where(number_of_zeros_in_row == min_zeros)[0]
        return indices_with_min_zeros_values

    @staticmethod
    def ordering_to_matrix_indices(path: np.ndarray, dtype=np.int16) -> np.ndarray:
        new_path = np.zeros(len(path), dtype=dtype)
        matrix_indices = np.arange(0, len(path), dtype=int)

        for i in range(len(path)):
            new_path[i] = matrix_indices[path[i]]
            matrix_indices = np.delete(matrix_indices, path[i])

        return new_path

    def __ordering_algorithm(self, array: np.array, path: list, all_paths: list,
                             max_index=None, use_only_first_path=False):
        if len(array) == 0:
            all_paths.append(path)
            return all_paths

        # 1) The program search for line numbers, witch contains desired conditions

        indexes = self._ordering_function(array, max_index=max_index)

        # 2) If we find desired rows we can reduce matrix

        if len(indexes) == 0:
            all_paths.append(path)

        if use_only_first_path:
            indexes = [indexes[0]]

        for index in indexes:
            reduced_array = np.delete(array, index, 0)
            reduced_array = np.delete(reduced_array, index, 1)

            new_path = path.copy()
            new_path.append(index)
            self.__ordering_algorithm(reduced_array, new_path, all_paths,
                                      max_index=max_index, use_only_first_path=use_only_first_path)

            if len(array) == max_index:
                break

        return all_paths


class NodeOrderingAlgorithm(OrderingAlgorithm):
    def _ordering_function(self, array: np.ndarray, max_index: int):
        return OrderingAlgorithm.find_indexes_with_min_zeros_count(array)

    def __str__(self):
        return "Node Ordering"


class SimpleSort(NodeOrderingAlgorithm):
    def __call__(self, sequence_array: np.ndarray, use_only_first_path=False):
        return self.__sort(sequence_array)

    def __sort(self, sequence_array: np.ndarray) -> list:
        only_zeros = np.where(sequence_array == 0, 1, 0)
        number_zeros_in_columns = np.sum(only_zeros, axis=1)

        sorted_indexes = [np.argsort(number_zeros_in_columns)]

        return sorted_indexes

    def __str__(self):
        return "Simple Sort"


class SimpleSortWeighted(NodeOrderingAlgorithm):
    def __call__(self, sequence_array: np.ndarray, use_only_first_path=False):
        return self.__sort(sequence_array)

    def __sort(self, sequence_array: np.ndarray) -> list:
        weight_row_sum = SimpleSortWeighted.create_wight_array_row_sum(sequence_array)

        only_zeros = np.where(sequence_array == 0, 1, 0)
        number_zeros_in_columns = np.sum(only_zeros, axis=1)

        important_weights = np.rec.fromarrays([number_zeros_in_columns, weight_row_sum])
        comparator = important_weights.astype(dtype=[('x', int), ('y', int)])

        sorted_indexes = [np.argsort(comparator, order=['x', 'y'])]

        return sorted_indexes

    @staticmethod
    def create_wight_array_row_sum(sequence_matrix: np.ndarray) -> np.ndarray:
        weight_array = sequence_matrix.copy()
        weight_array = EnhancedNodeOrderingAlgorithm.create_wight_array(weight_array)

        return np.sum(weight_array, axis=1)

    def __str__(self):
        return "Simple Sort Weighted"


class EnhancedNodeOrderingAlgorithm(OrderingAlgorithm):
    @staticmethod
    def create_wight_array(sequence_matrix: np.ndarray) -> np.ndarray:
        weight_array = sequence_matrix.copy().astype(dtype=float)

        with np.errstate(divide='ignore'):
            weight_array = np.where(weight_array == 0, weight_array, 1 / weight_array)

        return weight_array

    def _ordering_function(self, array: np.ndarray, max_index: int):
        indices_with_no_zeros = OrderingAlgorithm.find_indexes_with_min_zeros_count(array)

        if len(indices_with_no_zeros) == 0:
            return np.array([])

        weight_matrix = EnhancedNodeOrderingAlgorithm.create_wight_array(array)
        row_no_zeros_weight_sum = np.sum(weight_matrix[indices_with_no_zeros], axis=1)
        no_zeros_weight_sum = np.vstack((indices_with_no_zeros, row_no_zeros_weight_sum))

        min_indexes = np.where(no_zeros_weight_sum[1] == np.min(no_zeros_weight_sum[1]))[0]
        weight_min_indexed_array = no_zeros_weight_sum[:, min_indexes]
        weight_min_indexed_row = weight_min_indexed_array[0]

        return weight_min_indexed_row.astype(int)

    def __str__(self):
        return "Enhanced Node Ordering"


class WeightedEdgesNodeOrderingAlgorithm(OrderingAlgorithm):
    def _ordering_function(self, array: np.ndarray, max_index: int):
        indices_with_no_zeros = OrderingAlgorithm.find_indexes_with_min_zeros_count(array)

        if len(indices_with_no_zeros) == 0:
            return np.array([])

        correct_number = max_index - len(array)

        # !? this code can contain error

        max_index_array = indices_with_no_zeros

        if len(indices_with_no_zeros) > 1:
            max_correct_numbers = 0
            current_number_list = []

            for index in indices_with_no_zeros:
                current_correct_numbers = len(np.where(array[index] == correct_number)[0])

                if current_correct_numbers == max_correct_numbers:
                    current_number_list.append(index)

                if current_correct_numbers > max_correct_numbers:
                    current_number_list = [index]

                    max_correct_numbers = current_correct_numbers

            max_index_array = np.array(current_number_list).astype(int)

        return max_index_array

    def __str__(self):
        return "Weighted Edges Ordering"
